"""
Command line interface(s)
"""
import logging
import os
import os.path

import click

from .cli_utils import _get_timestamp, _init_logging, _make_path_if_not_exists
from .crunching import _crunch_data
from .definitions import OUTPUT_PREFIX
from .stitching import _stitch_netdf_scm_ncs
from .wrangling import _do_wrangling, _tuningstrucs_blended_model_wrangling

logger = logging.getLogger("netcdf_scm")


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("src", type=click.Path(exists=True, readable=True, resolve_path=True))
@click.argument(
    "dst", type=click.Path(file_okay=False, writable=True, resolve_path=True)
)
@click.argument("crunch_contact")
@click.option(
    "--drs",
    default="Scm",
    type=click.Choice(["Scm", "MarbleCMIP5", "CMIP6Input4MIPs", "CMIP6Output"]),
    show_default=True,
    help="Data reference syntax to use for crunching.",
)
@click.option(
    "--regexp",
    default="^(?!.*(fx)).*$",
    show_default=True,
    help=(
        "Regular expression to apply to file directory (only crunches "
        "matches). Be careful, if you use a very copmlex regexp directory "
        "sorting can be extremely slow (see e.g. discussion at "
        "https://stackoverflow.com/a/5428712)!"
    ),
)
@click.option(
    "--regions",
    default="World,World|Northern Hemisphere,World|Southern Hemisphere,World|Land,World|Ocean,World|Northern Hemisphere|Land,World|Southern Hemisphere|Land,World|Northern Hemisphere|Ocean,World|Southern Hemisphere|Ocean",
    show_default=True,
    help="Comma-separated regions to crunch.",
)
@click.option(
    "--data-sub-dir",
    default="{}-crunched".format(OUTPUT_PREFIX),
    show_default=True,
    help="Sub-directory of ``dst`` to save data in.",
)
@click.option(
    "--force/--do-not-force",
    "-f",
    help="Overwrite any existing files.",
    default=False,
    show_default=True,
)
@click.option(
    "--small-number-workers",
    default=10,
    show_default=True,
    help="Maximum number of workers to use when crunching files.",
)
@click.option(
    "--small-threshold",
    default=50.0,
    show_default=True,
    help="Maximum number of data points (in millions) in a file for it to be processed in parallel with ``small-number-workers``",
)
@click.option(
    "--medium-number-workers",
    default=3,
    show_default=True,
    help="Maximum number of workers to use when crunching files.",
)
@click.option(
    "--medium-threshold",  # pylint:disable=too-many-arguments,too-many-locals
    default=120.0,
    show_default=True,
    help="Maximum number of data points (in millions) in a file for it to be processed in parallel with ``medium-number-workers``",
)
@click.option(
    "--force-lazy-threshold",
    default=1000.0,
    show_default=True,
    help="Maximum number of data points (in millions) in a file for it to be processed in memory",
)
@click.option(
    "--cell-weights",
    default=None,
    type=click.Choice(["area-only", "area-surface-fraction"]),
    show_default=True,
    help=(
        "How to weight cells when calculating aggregates. "
        "If 'area-surface-fraction', land surface fraction weights will be included when taking cell means. "
        "If 'area-only', land surface fraction weights will not be included when taking cell means, "
        "hence cells will only be weighted by their area. "
        "If nothing is provided, netCDF-SCM will guess whether land surface fraction weights should "
        "be included or not based on the data being processed. "
        "See :meth:`netcdf_scm.iris_cube_wrappers.ScmCube.get_scm_timeseries_weights` for more details."
    ),
)
def crunch_data(
    src,
    dst,
    crunch_contact,
    drs,
    regexp,
    regions,
    data_sub_dir,
    force,
    small_number_workers,
    small_threshold,
    medium_number_workers,
    medium_threshold,
    force_lazy_threshold,
    cell_weights,
):
    r"""
    Crunch data in ``src`` to netCDF-SCM ``.nc`` files in ``dst``.

    ``src`` is searched recursively and netcdf-scm will attempt to crunch all the files
    found. The directory structure in ``src`` will be mirrored in ``dst``.

    Failures and warnings are recorded and written into a text file in ``dst``. We
    recommend examining this file using a file analysis tool such as ``grep``. We
    often use the command ``grep "\|WARNING\|INFO\|ERROR <log-file>``.

    ``crunch_contact`` is written into the output ``.nc`` files' ``crunch_contact``
    attribute.
    """
    _crunch_data(
        logger,
        src,
        dst,
        crunch_contact,
        drs,
        regexp,
        regions,
        data_sub_dir,
        force,
        small_number_workers,
        small_threshold,
        medium_number_workers,
        medium_threshold,
        force_lazy_threshold,
        cell_weights,
    )


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("src", type=click.Path(exists=True, readable=True, resolve_path=True))
@click.argument(
    "dst", type=click.Path(file_okay=False, writable=True, resolve_path=True)
)
@click.argument("wrangle_contact")
@click.option(
    "--regexp",
    default="^(?!.*(fx)).*$",
    show_default=True,
    help=(
        "Regular expression to apply to file directory (only wrangles "
        "matches). Be careful, if you use a very copmlex regexp directory "
        "sorting can be extremely slow (see e.g. discussion at "
        "https://stackoverflow.com/a/5428712)!"
    ),
)
@click.option(
    "--prefix", default=None, help="Prefix to apply to output file names (not paths)."
)
@click.option(
    "--out-format",
    default="mag-files",
    type=click.Choice(
        [
            "mag-files",
            "mag-files-average-year-start-year",
            "mag-files-average-year-mid-year",
            "mag-files-average-year-end-year",
            "mag-files-point-start-year",
            "mag-files-point-mid-year",
            "mag-files-point-end-year",
            "magicc-input-files",
            "magicc-input-files-average-year-start-year",
            "magicc-input-files-average-year-mid-year",
            "magicc-input-files-average-year-end-year",
            "magicc-input-files-point-start-year",
            "magicc-input-files-point-mid-year",
            "magicc-input-files-point-end-year",
            "tuningstrucs-blend-model",
        ]
    ),
    show_default=True,
    help=(
        "Format to re-write crunched data into. The time operation conventions follow "
        "those in "
        "`Pymagicc <https://pymagicc.readthedocs.io/en/latest/file_conventions.html#namelists>`_."
    ),
)
@click.option(
    "--drs",
    default="None",
    type=click.Choice(["None", "MarbleCMIP5", "CMIP6Input4MIPs", "CMIP6Output"]),
    show_default=True,
    help="Data reference syntax to use to decipher paths. This is required to ensure the output folders match the input data reference syntax.",
)
@click.option(
    "--force/--do-not-force",
    "-f",
    help="Overwrite any existing files.",
    default=False,
    show_default=True,
)  # pylint:disable=too-many-arguments
@click.option(
    "--number-workers",  # pylint:disable=too-many-arguments
    help="Number of worker (threads) to use when wrangling.",
    default=4,
    show_default=True,
)
@click.option(
    "--target-units-specs",  # pylint:disable=too-many-arguments
    help="csv containing target units for wrangled variables.",
    default=None,
    show_default=False,
    type=click.Path(exists=True, readable=True, resolve_path=True),
)
def wrangle_netcdf_scm_ncs(
    src,
    dst,
    wrangle_contact,
    regexp,
    prefix,
    out_format,
    drs,
    force,
    number_workers,
    target_units_specs,
):
    """
    Wrangle netCDF-SCM ``.nc`` files into other formats and directory structures.

    ``src`` is searched recursively and netcdf-scm will attempt to wrangle all the
    files found.

    ``wrangle_contact`` is written into the header of the output files.
    """
    log_file = os.path.join(
        dst,
        "{}-wrangle.log".format(_get_timestamp().replace(" ", "_").replace(":", "")),
    )
    _make_path_if_not_exists(dst)
    _init_logging(
        logger,
        [
            ("wrangle_contact", wrangle_contact),
            ("source", src),
            ("destination", dst),
            ("regexp", regexp),
            ("prefix", prefix),
            ("drs", drs),
            ("out_format", out_format),
            ("force", force),
        ],
        out_filename=log_file,
    )

    # TODO: turn all wranglers into subclasses of a base `Wrangler` class
    if out_format == "tuningstrucs-blend-model":
        _tuningstrucs_blended_model_wrangling(
            logger, src, dst, regexp, force, drs, prefix
        )
    else:
        _do_wrangling(
            logger,
            src,
            dst,
            regexp,
            out_format,
            force,
            wrangle_contact,
            drs,
            number_workers,
            target_units_specs,
        )


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("src", type=click.Path(exists=True, readable=True, resolve_path=True))
@click.argument(
    "dst", type=click.Path(file_okay=False, writable=True, resolve_path=True)
)
@click.argument("stitch_contact")
@click.option(
    "--regexp",
    default="^(?!.*(fx)).*$",
    show_default=True,
    help=(
        "Regular expression to apply to file directory (only stitches "
        "matches). Be careful, if you use a very copmlex regexp directory "
        "sorting can be extremely slow (see e.g. discussion at "
        "https://stackoverflow.com/a/5428712)!"
    ),
)
@click.option(
    "--prefix", default=None, help="Prefix to apply to output file names (not paths)."
)
@click.option(
    "--out-format",
    default="mag-files",
    type=click.Choice(
        [
            "mag-files",
            "mag-files-average-year-start-year",
            "mag-files-average-year-mid-year",
            "mag-files-average-year-end-year",
            "mag-files-point-start-year",
            "mag-files-point-mid-year",
            "mag-files-point-end-year",
            "magicc-input-files",
            "magicc-input-files-average-year-start-year",
            "magicc-input-files-average-year-mid-year",
            "magicc-input-files-average-year-end-year",
            "magicc-input-files-point-start-year",
            "magicc-input-files-point-mid-year",
            "magicc-input-files-point-end-year",
            "tuningstrucs-blend-model",
        ]
    ),
    show_default=True,
    help=(
        "Format to re-write crunched data into. The time operation conventions follow "
        "those in "
        "`Pymagicc <https://pymagicc.readthedocs.io/en/latest/file_conventions.html#namelists>`_ ."
    ),
)
@click.option(
    "--drs",
    default="None",
    type=click.Choice(["None", "MarbleCMIP5", "CMIP6Input4MIPs", "CMIP6Output"]),
    show_default=True,
    help="Data reference syntax to use to decipher paths. This is required to ensure the output folders match the input data reference syntax.",
)
@click.option(
    "--force/--do-not-force",
    "-f",
    help="Overwrite any existing files.",
    default=False,
    show_default=True,
)  # pylint:disable=too-many-arguments
@click.option(
    "--number-workers",  # pylint:disable=too-many-arguments
    help="Number of worker (threads) to use when stitching.",
    default=4,
    show_default=True,
)
@click.option(
    "--target-units-specs",  # pylint:disable=too-many-arguments
    help="csv containing target units for stitched variables.",
    default=None,
    show_default=False,
    type=click.Path(exists=True, readable=True, resolve_path=True),
)
@click.option(
    "--normalise",
    default=None,
    type=click.Choice(
        [
            "31-yr-mean-after-branch-time",
            "21-yr-running-mean",
            "21-yr-running-mean-dedrift",
        ]
    ),
    show_default=False,
    help="How to normalise the data relative to piControl (if not provided, no normalisation is performed).",
)
def stitch_netcdf_scm_ncs(
    src,
    dst,
    stitch_contact,
    regexp,
    prefix,
    out_format,
    drs,
    force,
    number_workers,
    target_units_specs,
    normalise,
):
    """
    Stitch netCDF-SCM ``.nc`` files together and write out in the specified format.

    ``SRC`` is searched recursively and netcdf-scm will attempt to stitch all the
    files found. Output is written in ``DST``.

    ``STITCH_CONTACT`` is written into the header of the output files.
    """
    log_file = os.path.join(
        dst,
        "{}-stitch.log".format(_get_timestamp().replace(" ", "_").replace(":", "")),
    )
    _make_path_if_not_exists(dst)
    _init_logging(
        logger,
        [
            ("stitch-contact", stitch_contact),
            ("source", src),
            ("destination", dst),
            ("regexp", regexp),
            ("prefix", prefix),
            ("out-format", out_format),
            ("drs", drs),
            ("force", force),
            ("number-workers", number_workers),
            ("target-units-specs", target_units_specs),
            ("normalise", normalise),
        ],
        out_filename=log_file,
    )

    _stitch_netdf_scm_ncs(
        logger,
        src,
        dst,
        stitch_contact,
        regexp,
        prefix,
        out_format,
        drs,
        force,
        number_workers,
        target_units_specs,
        normalise,
    )



"""
Helper tools for citing CMIP6 of the `Coupled Model Intercomparison Project <https://www.wcrp-climate.org/wgcm-cmip>`_ data
"""
import datetime as dt
import functools
import json

import requests

from ..errors import NoLicenseInformationError, NonStandardLicenseError


def check_license_cmip6(scmdata):
    """
    Check the license information of a dataset

    Parameters
    ----------
    scmdata : :obj:`scmdata.ScmRun`
        Dataset of which to check the license information

    Raises
    ------
    NonStandardLicenseError
        The dataset's license is non-standard i.e. is not a
        `Creative Commons Attribution ShareAlike 4.0 International License <https://creativecommons.org/licenses/by-sa/4.0/>`_

    NoLicenseInformationError
        No license information could be found in ``scmdata.metadata``
    """
    licenses = get_license_info(scmdata)

    non_ccas4 = {
        license
        for license in licenses
        if not (
            ("Creative Commons Attribution" in license)
            and ("ShareAlike 4.0 International License" in license)
            and (
                "for terms of use governing CMIP6 output, including citation requirements and proper acknowledgment"
                in license
            )
            and ("NonCommercial" not in license)
        )
    }
    if non_ccas4:
        error_msg = "Non-standard licenses: {}".format(non_ccas4)
        raise NonStandardLicenseError(error_msg)


def get_license_info(scmdata):
    """
    Get the license information from a dataset's metadata.

    Parameters
    ----------
    scmdata : :obj:`scmdata.ScmRun`
        Dataset from which to get the license information

    Raises
    ------
    NoLicenseInformationError
        No license information could be found in ``scmdata.metadata``
    """
    licenses = [v for k, v in scmdata.metadata.items() if "license" in k]

    if not licenses:
        raise NoLicenseInformationError("Metadata found: {}".format(scmdata.metadata))

    return licenses


@functools.lru_cache()
def _get_cmip6_doi(drs_id):  # pragma: no cover
    params = (
        ("input", drs_id),
        ("wt", "json"),
    )
    response = requests.get(
        "https://cera-www.dkrz.de/WDCC/ui/cerasearch/cerarest/exportcmip6",
        params=params,
    )
    response.raise_for_status()

    cite_info = json.loads(response.text)
    doi = cite_info["identifier"]["id"]

    return doi


@functools.lru_cache()
def _query_cmip6(query_url):  # pragma: no cover
    headers = {
        "Accept": "application/x-bibtex",
    }

    response = requests.get(query_url, headers=headers)
    response.raise_for_status()

    return response.text


def get_cmip6_bibtex_citation(
    mip_era, activity_id, institute_id, model, scenario, version=None
):
    """
    Get CMIP5 bibtex citation

    Parameters
    ----------
    mip_era : str
        Mip era (should always be CMIP6)

    activity_id : str
        Activity id (e.g. ScenarioMIP)

    institute_id : str
        Institute which produced the output

    model : str
        Model of interest

    scenario : str
        Scenario of interest

    version : Union[str, None]
        Version of the dataset. If ``None``, today's date is used in accordance with the guidance provided by the CMIP6 citation guidelines
        `see e.g. footnote 1 here <https://cera-www.dkrz.de/WDCC/ui/cerasearch/cmip6?input=CMIP6.ScenarioMIP.CSIRO-ARCCSS.ACCESS-CM2.ssp585>`_

    Returns
    -------
    str
        Bibtex citation

    Raises
    ------
    ValueError
        No CMIP5 data is available for ``model`` ``scenario`` hence no
        citation can be provided.
    """
    drs_id = ".".join([mip_era, activity_id, institute_id, model, scenario])

    citation_key = _get_cmip6_citation_key(
        mip_era, activity_id, institute_id, model, scenario
    )
    try:
        doi = _get_cmip6_doi(drs_id)
        query_url = "https://doi.org/{}".format(doi)
        bib_tex_citation = _query_cmip6(query_url).replace(
            query_url.lower(), citation_key
        )

    except requests.exceptions.HTTPError as exc:
        raise ValueError(
            "No citation could be found for {}.\nError was: {}".format(drs_id, exc)
        )

    if version is None:
        version = dt.date.today().strftime("%Y%m%d")

    title = _get_title(bib_tex_citation)
    bib_tex_citation = bib_tex_citation.replace(
        title, "{}. Version {}".format(title, version)
    )

    return bib_tex_citation


def _get_title(bib_tex_citation):
    title = bib_tex_citation.split("title")[1].split("{")[1].split("}")[0]

    return title


def _get_cmip6_citation_key(mip_era, activity_id, institute_id, model, scenario):
    return "cmip6data_{}".format(
        "_".join([mip_era, activity_id, institute_id, model, scenario])
    )



"""
Test that crunching timeseries behaves as expected using files we have written

Helpful snippets:
python -c "from test_scm_timeseries_crunching import write_test_files; wp = '.'; write_test_files(wp)"

from matplotlib import pyplot as plt
import iris.quickplot as qplt
qplt.pcolor(default_sftlf_cube); plt.gca().coastlines(); plt.show()  # maps
qplt.pcolor(cube); plt.gca().coastlines(); plt.show()  # maps

qplt.plot(cube[:, 1, 3]); plt.show()  # timeseries
"""
import datetime as dt
import json
import os.path
import re

import iris
import numpy as np
import pytest
from scmdata import ScmDataFrame

from netcdf_scm.iris_cube_wrappers import MarbleCMIP5Cube

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

RAW_DATA = np.array(
    [
        [[30, 40, 50, 60], [110, 120, 190, 260], [3, 60, 20, 40]],
        [[0, 15, 45, 90], [300, 350, 450, 270], [10, 70, 90, 130]],
        [[60, 120, 60, 60], [510, 432, 220, 280], [50, 60, 55, 60]],
    ]
)

SURFACE_FRACS = np.array([[0, 30, 0, 10], [80, 100, 0, 50], [20, 10, 51, 15]])
AREA_WEIGHTS = np.array([[1.2, 1.2, 1.2, 1.2], [2, 2, 2, 2], [1.1, 1.1, 1.1, 1.1]])
WORLD_AREA = 510 * 10 ** 12
AREA_SCALE_FACTOR = WORLD_AREA / np.sum(AREA_WEIGHTS)
AREA_WEIGHTS = AREA_WEIGHTS * AREA_SCALE_FACTOR

TEST_RSDT_PATH = os.path.join(
    ROOT_DIR,
    "cmip5",
    "experiment",
    "Amon",
    "rsdt",
    "model",
    "realisation",
    "rsdt_Amon_model_experiment_realisation_185001-185003.nc",
)
TEST_GPP_PATH = os.path.join(
    ROOT_DIR,
    "cmip5",
    "experiment",
    "Lmon",
    "gpp",
    "model",
    "realisation",
    "gpp_Lmon_model_experiment_realisation_185001-185003.nc",
)
TEST_CSOILFAST_PATH = os.path.join(
    ROOT_DIR,
    "cmip5",
    "experiment",
    "Lmon",
    "cSoilFast",
    "model",
    "realisation",
    "cSoilFast_Lmon_model_experiment_realisation_185001-185003.nc",
)
TEST_HFDS_PATH = os.path.join(
    ROOT_DIR,
    "cmip5",
    "experiment",
    "Omon",
    "hfds",
    "model",
    "realisation",
    "hfds_Omon_model_experiment_realisation_185001-185003.nc",
)

TEST_AREACEALLA_PATH = os.path.join(
    ROOT_DIR,
    "cmip5",
    "experiment",
    "fx",  # TODO: Be careful when writing table id retrieval test as this is Ofx for CMIP6Output
    "areacella",
    "model",
    "r0i0p0",
    "areacella_fx_model_experiment_r0i0p0.nc",
)
TEST_AREACEALLO_PATH = os.path.join(
    ROOT_DIR,
    "cmip5",
    "experiment",
    "fx",
    "areacello",
    "model",
    "r0i0p0",
    "areacello_fx_model_experiment_r0i0p0.nc",
)

TEST_SFTLF_PATH = os.path.join(
    ROOT_DIR,
    "cmip5",
    "experiment",
    "fx",
    "sftlf",
    "model",
    "r0i0p0",
    "sftlf_fx_model_experiment_r0i0p0.nc",
)
TEST_SFTOF_PATH = os.path.join(
    ROOT_DIR,
    "cmip5",
    "experiment",
    "fx",
    "sftof",
    "model",
    "r0i0p0",
    "sftof_fx_model_experiment_r0i0p0.nc",
)

SCMDF_TIME = [
    dt.datetime(1850, 1, 16, 12),
    dt.datetime(1850, 2, 15, 0),
    dt.datetime(1850, 3, 16, 12),
]


LAND_EFFECTIVE_AREAS = (SURFACE_FRACS * AREA_WEIGHTS) / 100
OCEAN_EFFECTIVE_AREAS = ((100 - SURFACE_FRACS) * AREA_WEIGHTS) / 100
AREAS = {
    "World": np.sum(AREA_WEIGHTS),
    "World|Land": np.sum(LAND_EFFECTIVE_AREAS),
    "World|Ocean": np.sum(OCEAN_EFFECTIVE_AREAS),
    "World|Northern Hemisphere": np.sum(AREA_WEIGHTS[:2, :]),
    "World|Southern Hemisphere": np.sum(AREA_WEIGHTS[2, :]),
    "World|Northern Hemisphere|Land": np.sum(LAND_EFFECTIVE_AREAS[:2, :]),
    "World|Southern Hemisphere|Land": np.sum(LAND_EFFECTIVE_AREAS[2, :]),
    "World|Northern Hemisphere|Ocean": np.sum(OCEAN_EFFECTIVE_AREAS[:2, :]),
    "World|Southern Hemisphere|Ocean": np.sum(OCEAN_EFFECTIVE_AREAS[2, :]),
    "World|North Atlantic Ocean": 1 * AREA_SCALE_FACTOR,
    "World|El Nino N3.4": 2 * AREA_SCALE_FACTOR,
}
OCEAN_MASK = ((100 - SURFACE_FRACS) > 0).astype(float)
OCEAN_MASK_EFFECTIVE_AREAS = AREA_WEIGHTS * OCEAN_MASK
AREAS_OCEAN_MASK = {
    "World": np.sum(AREA_WEIGHTS),
    "World|Ocean": np.sum(OCEAN_MASK_EFFECTIVE_AREAS),
    "World|Northern Hemisphere": np.sum(OCEAN_MASK_EFFECTIVE_AREAS[:2, :]),
    "World|Southern Hemisphere": np.sum(OCEAN_MASK_EFFECTIVE_AREAS[2, :]),
    "World|Northern Hemisphere|Ocean": np.sum(OCEAN_MASK_EFFECTIVE_AREAS[:2, :]),
    "World|Southern Hemisphere|Ocean": np.sum(OCEAN_MASK_EFFECTIVE_AREAS[2, :]),
    "World|North Atlantic Ocean": 2 * AREA_SCALE_FACTOR,
    "World|El Nino N3.4": 2 * AREA_SCALE_FACTOR,
}


def _add_land_area_metadata(in_scmdf, realm):
    if realm == "ocean":
        areas = AREAS_OCEAN_MASK
    else:
        areas = AREAS

    for region in in_scmdf["region"].unique():
        r_key = "area_{} (m**2)".format(
            region.lower().replace("|", "_").replace(" ", "_")
        )
        if region in (
            "World",
            "World|Northern Hemisphere",
            "World|Southern Hemisphere",
        ):
            if realm in ("land", "ocean"):
                in_scmdf.metadata[r_key] = areas[
                    "{}|{}".format(region, realm.capitalize())
                ]
                continue

        in_scmdf.metadata[r_key] = areas[region]

    return in_scmdf


def get_rsdt_expected_results():
    world_values = np.sum(np.sum(RAW_DATA * AREA_WEIGHTS, axis=2), axis=1) / np.sum(
        AREA_WEIGHTS
    )

    land_weights = SURFACE_FRACS * AREA_WEIGHTS
    world_land_values = np.sum(
        np.sum(RAW_DATA * land_weights, axis=2), axis=1
    ) / np.sum(land_weights)

    ocean_weights = (100 - SURFACE_FRACS) * AREA_WEIGHTS / 100
    world_ocean_values = np.sum(
        np.sum(RAW_DATA * ocean_weights, axis=2), axis=1
    ) / np.sum(ocean_weights)

    nh_area_weights = np.copy(AREA_WEIGHTS)
    nh_area_weights[2, :] = 0
    world_nh_values = np.sum(
        np.sum(RAW_DATA * nh_area_weights, axis=2), axis=1
    ) / np.sum(nh_area_weights)

    sh_area_weights = np.copy(AREA_WEIGHTS)
    sh_area_weights[:2, :] = 0
    world_sh_values = np.sum(
        np.sum(RAW_DATA * sh_area_weights, axis=2), axis=1
    ) / np.sum(sh_area_weights)

    # we do these by hand: yes they're very slow but that's the point
    world_nh_land_values = np.array(
        [
            (40 * 30 + 60 * 10) * 1.2 + (110 * 80 + 120 * 100 + 260 * 50) * 2,
            (15 * 30 + 90 * 10) * 1.2 + (300 * 80 + 350 * 100 + 270 * 50) * 2,
            (120 * 30 + 60 * 10) * 1.2 + (510 * 80 + 432 * 100 + 280 * 50) * 2,
        ]
    ) / ((30 + 10) * 1.2 + (80 + 100 + 50) * 2)

    world_sh_land_values = np.array(
        [
            (3 * 20 + 60 * 10 + 20 * 51 + 40 * 15) * 1.1,
            (10 * 20 + 70 * 10 + 90 * 51 + 130 * 15) * 1.1,
            (50 * 20 + 60 * 10 + 55 * 51 + 60 * 15) * 1.1,
        ]
    ) / ((20 + 10 + 51 + 15) * 1.1)

    world_nh_ocean_values = np.array(
        [
            (30 * 100 + 40 * 70 + 50 * 100 + 60 * 90) * 1.2
            + (110 * 20 + 190 * 100 + 260 * 50) * 2,
            (0 * 100 + 15 * 70 + 45 * 100 + 90 * 90) * 1.2
            + (300 * 20 + 450 * 100 + 270 * 50) * 2,
            (60 * 100 + 120 * 70 + 60 * 100 + 60 * 90) * 1.2
            + (510 * 20 + 220 * 100 + 280 * 50) * 2,
        ]
    ) / ((100 + 70 + 100 + 90) * 1.2 + (20 + 100 + 50) * 2)

    world_sh_ocean_values = np.array(
        [
            (3 * 80 + 60 * 90 + 20 * 49 + 40 * 85) * 1.1,
            (10 * 80 + 70 * 90 + 90 * 49 + 130 * 85) * 1.1,
            (50 * 80 + 60 * 90 + 55 * 49 + 60 * 85) * 1.1,
        ]
    ) / ((80 + 90 + 49 + 85) * 1.1)

    world_north_atlantic_values = np.array([260, 270, 280])

    world_elnino_values = np.array([190, 450, 220])

    data = np.vstack(
        [
            world_values,
            world_land_values,
            world_ocean_values,
            world_nh_values,
            world_sh_values,
            world_nh_land_values,
            world_sh_land_values,
            world_nh_ocean_values,
            world_sh_ocean_values,
            world_north_atlantic_values,
            world_elnino_values,
        ]
    ).T

    exp = ScmDataFrame(
        data=data,
        index=SCMDF_TIME,
        columns={
            "model": "unspecified",
            "scenario": "experiment",
            "region": [
                "World",
                "World|Land",
                "World|Ocean",
                "World|Northern Hemisphere",
                "World|Southern Hemisphere",
                "World|Northern Hemisphere|Land",
                "World|Southern Hemisphere|Land",
                "World|Northern Hemisphere|Ocean",
                "World|Southern Hemisphere|Ocean",
                "World|North Atlantic Ocean",
                "World|El Nino N3.4",
            ],
            "variable": "rsdt",
            "unit": "W m^-2",
            "climate_model": "model",
            "activity_id": "cmip5",
            "member_id": "realisation",
            "variable_standard_name": "toa_incoming_shortwave_flux",
            "mip_era": "CMIP5",
        },
    )

    exp.metadata = {
        "calendar": "gregorian",
        "land_fraction": np.sum(AREA_WEIGHTS * SURFACE_FRACS)
        / (100 * np.sum(AREA_WEIGHTS)),
        "land_fraction_northern_hemisphere": np.sum(nh_area_weights * SURFACE_FRACS)
        / (100 * np.sum(nh_area_weights)),
        "land_fraction_southern_hemisphere": np.sum(sh_area_weights * SURFACE_FRACS)
        / (100 * np.sum(sh_area_weights)),
        "modeling_realm": "atmos",
        "Conventions": "CF-1.7",
        "crunch_source_files": "Files: ['/cmip5/experiment/Amon/rsdt/model/realisation/rsdt_Amon_model_experiment_realisation_185001-185003.nc']; areacella: ['/cmip5/experiment/fx/areacella/model/r0i0p0/areacella_fx_model_experiment_r0i0p0.nc']; sftlf: ['/cmip5/experiment/fx/sftlf/model/r0i0p0/sftlf_fx_model_experiment_r0i0p0.nc']",
    }
    exp = _add_land_area_metadata(exp, realm="atmos")

    return exp


def get_gpp_expected_results():
    land_weights = SURFACE_FRACS * AREA_WEIGHTS
    world_values = np.sum(np.sum(RAW_DATA * land_weights, axis=2), axis=1) / np.sum(
        land_weights
    )

    world_land_values = world_values

    nh_area_weights = np.copy(AREA_WEIGHTS)
    nh_area_weights[2, :] = 0
    # we do these by hand: yes they're very slow but that's the point
    world_nh_land_values = np.array(
        [
            (40 * 30 + 60 * 10) * 1.2 + (110 * 80 + 120 * 100 + 260 * 50) * 2,
            (15 * 30 + 90 * 10) * 1.2 + (300 * 80 + 350 * 100 + 270 * 50) * 2,
            (120 * 30 + 60 * 10) * 1.2 + (510 * 80 + 432 * 100 + 280 * 50) * 2,
        ]
    ) / ((30 + 10) * 1.2 + (80 + 100 + 50) * 2)
    world_nh_values = world_nh_land_values

    sh_area_weights = np.copy(AREA_WEIGHTS)
    sh_area_weights[:2, :] = 0
    world_sh_land_values = np.array(
        [
            (3 * 20 + 60 * 10 + 20 * 51 + 40 * 15) * 1.1,
            (10 * 20 + 70 * 10 + 90 * 51 + 130 * 15) * 1.1,
            (50 * 20 + 60 * 10 + 55 * 51 + 60 * 15) * 1.1,
        ]
    ) / ((20 + 10 + 51 + 15) * 1.1)
    world_sh_values = world_sh_land_values

    data = np.vstack(
        [
            world_values,
            world_land_values,
            world_nh_values,
            world_sh_values,
            world_nh_land_values,
            world_sh_land_values,
        ]
    ).T

    exp = ScmDataFrame(
        data=data,
        index=SCMDF_TIME,
        columns={
            "model": "unspecified",
            "scenario": "experiment",
            "region": [
                "World",
                "World|Land",
                "World|Northern Hemisphere",
                "World|Southern Hemisphere",
                "World|Northern Hemisphere|Land",
                "World|Southern Hemisphere|Land",
            ],
            "variable": "gpp",
            "unit": "kg m^-2 s^-1",
            "climate_model": "model",
            "activity_id": "cmip5",
            "member_id": "realisation",
            "variable_standard_name": "gross_primary_productivity_of_carbon",
            "mip_era": "CMIP5",
        },
    )
    exp.metadata = {
        "calendar": "gregorian",
        "modeling_realm": "land",
        "Conventions": "CF-1.7",
        "crunch_source_files": "Files: ['/cmip5/experiment/Lmon/gpp/model/realisation/gpp_Lmon_model_experiment_realisation_185001-185003.nc']; sftlf: ['/cmip5/experiment/fx/sftlf/model/r0i0p0/sftlf_fx_model_experiment_r0i0p0.nc']; areacella: ['/cmip5/experiment/fx/areacella/model/r0i0p0/areacella_fx_model_experiment_r0i0p0.nc']",
    }
    exp = _add_land_area_metadata(exp, realm="land")

    return exp


def get_csoilfast_expected_results():
    exp_scmdf = get_gpp_expected_results()
    exp_scmdf.set_meta("cSoilFast", "variable")
    exp_scmdf.set_meta("fast_soil_pool_carbon_content", "variable_standard_name")
    exp_scmdf.set_meta("kg m^-2", "unit")

    exp_scmdf.metadata[
        "crunch_source_files"
    ] = "Files: ['/cmip5/experiment/Lmon/cSoilFast/model/realisation/cSoilFast_Lmon_model_experiment_realisation_185001-185003.nc']; sftlf: ['/cmip5/experiment/fx/sftlf/model/r0i0p0/sftlf_fx_model_experiment_r0i0p0.nc']; areacella: ['/cmip5/experiment/fx/areacella/model/r0i0p0/areacella_fx_model_experiment_r0i0p0.nc']"
    exp_scmdf = _add_land_area_metadata(exp_scmdf, realm="land")

    return exp_scmdf


def get_hfds_expected_results():
    ocean_weights = AREA_WEIGHTS * OCEAN_MASK
    world_values = np.sum(np.sum(RAW_DATA * ocean_weights, axis=2), axis=1) / np.sum(
        ocean_weights
    )

    world_ocean_values = world_values

    nh_area_weights = np.copy(AREA_WEIGHTS)
    nh_area_weights[2, :] = 0
    # we do these by hand: yes they're very slow but that's the point
    # hfds is area weighted only because it's ocean
    world_nh_ocean_values = np.array(
        [
            (30 + 40 + 50 + 60) * 1.2 + (110 + 190 + 260) * 2,
            (0 + 15 + 45 + 90) * 1.2 + (300 + 450 + 270) * 2,
            (60 + 120 + 60 + 60) * 1.2 + (510 + 220 + 280) * 2,
        ]
    ) / (4 * 1.2 + 3 * 2)
    world_nh_values = world_nh_ocean_values

    sh_area_weights = np.copy(AREA_WEIGHTS)
    sh_area_weights[:2, :] = 0
    world_sh_ocean_values = np.array(
        [
            (3 + 60 + 20 + 40) * 1.1,
            (10 + 70 + 90 + 130) * 1.1,
            (50 + 60 + 55 + 60) * 1.1,
        ]
    ) / (4 * 1.1)
    world_sh_values = world_sh_ocean_values

    world_north_atlantic_values = np.array([260, 270, 280])

    world_elnino_values = np.array([190, 450, 220])

    data = np.vstack(
        [
            world_values,
            world_ocean_values,
            world_nh_values,
            world_sh_values,
            world_nh_ocean_values,
            world_sh_ocean_values,
            world_north_atlantic_values,
            world_elnino_values,
        ]
    ).T

    exp = ScmDataFrame(
        data=data,
        index=SCMDF_TIME,
        columns={
            "model": "unspecified",
            "scenario": "experiment",
            "region": [
                "World",
                "World|Ocean",
                "World|Northern Hemisphere",
                "World|Southern Hemisphere",
                "World|Northern Hemisphere|Ocean",
                "World|Southern Hemisphere|Ocean",
                "World|North Atlantic Ocean",
                "World|El Nino N3.4",
            ],
            "variable": "hfds",
            "unit": "W m^-2",
            "climate_model": "model",
            "activity_id": "cmip5",
            "member_id": "realisation",
            "variable_standard_name": "surface_downward_heat_flux_in_sea_water",
            "mip_era": "CMIP5",
        },
    )
    exp.metadata = {
        "calendar": "gregorian",
        "modeling_realm": "ocean",
        "Conventions": "CF-1.7",
        "crunch_source_files": "Files: ['/cmip5/experiment/Omon/hfds/model/realisation/hfds_Omon_model_experiment_realisation_185001-185003.nc']; sftof: ['/cmip5/experiment/fx/sftof/model/r0i0p0/sftof_fx_model_experiment_r0i0p0.nc']; areacello: ['/cmip5/experiment/fx/areacello/model/r0i0p0/areacello_fx_model_experiment_r0i0p0.nc']",
    }
    exp = _add_land_area_metadata(exp, realm="ocean")

    return exp


@pytest.mark.parametrize(
    "test_data,invalid_regions,expected_results",
    [
        (TEST_RSDT_PATH, None, get_rsdt_expected_results()),
        (
            TEST_GPP_PATH,
            {
                "World|Ocean",
                "World|Northern Hemisphere|Ocean",
                "World|Southern Hemisphere|Ocean",
                "World|North Atlantic Ocean",
                "World|El Nino N3.4",
            },
            get_gpp_expected_results(),
        ),
        (
            TEST_CSOILFAST_PATH,
            {
                "World|Ocean",
                "World|Northern Hemisphere|Ocean",
                "World|Southern Hemisphere|Ocean",
                "World|North Atlantic Ocean",
                "World|El Nino N3.4",
            },
            get_csoilfast_expected_results(),
        ),
        (
            TEST_HFDS_PATH,
            {
                "World|Land",
                "World|Northern Hemisphere|Land",
                "World|Southern Hemisphere|Land",
            },
            get_hfds_expected_results(),
        ),
    ],
)
def test_scm_timeseries_crunching(
    assert_scmdata_frames_allclose, test_data, invalid_regions, expected_results
):
    tcube = MarbleCMIP5Cube()
    tcube.load_data_from_path(test_data)
    all_regions = {
        "World",
        "World|Land",
        "World|Ocean",
        "World|Northern Hemisphere",
        "World|Southern Hemisphere",
        "World|Northern Hemisphere|Land",
        "World|Southern Hemisphere|Land",
        "World|Northern Hemisphere|Ocean",
        "World|Southern Hemisphere|Ocean",
        "World|North Atlantic Ocean",
        "World|El Nino N3.4",
    }
    regions = (
        list(all_regions - invalid_regions)
        if invalid_regions is not None
        else list(all_regions)
    )
    if invalid_regions is not None:
        for r in invalid_regions:
            error_msg = re.escape("All weights are zero for region: `{}`".format(r))
            with pytest.raises(ValueError, match=error_msg):
                tcube.get_scm_timeseries(regions=[r])

    res = tcube.get_scm_timeseries(regions=regions)
    expected_results.metadata["crunch_netcdf_scm_weight_kwargs"] = json.dumps(
        {"regions": regions}
    )
    assert_scmdata_frames_allclose(res, expected_results)


def test_scm_timeseries_crunching_atmos_area_only():
    tcube = MarbleCMIP5Cube()
    tcube.load_data_from_path(TEST_RSDT_PATH)
    regions = [
        "World",
        "World|Land",
        "World|Ocean",
        "World|Northern Hemisphere",
        "World|Southern Hemisphere",
        "World|Northern Hemisphere|Land",
        "World|Southern Hemisphere|Land",
        "World|Northern Hemisphere|Ocean",
        "World|Southern Hemisphere|Ocean",
        "World|North Atlantic Ocean",
        "World|El Nino N3.4",
    ]

    error_msg = (
        re.escape("World|Land area of 379.535 million km^2 is clearly wrong") + ".*"
    )
    with pytest.raises(ValueError, match=error_msg):
        # If we crunch like this, then we give full area weights to cells
        # which are neither fully ocean nor fully land. Hence we double count
        # those cells areas and so the total World area is not equal to the
        # sum of land and ocean.
        # At some point this should probably be better explained and
        # documented, for now this will do.
        tcube.get_scm_timeseries(regions=regions, cell_weights="area-only")


def write_test_files(write_path):
    lat = iris.coords.DimCoord(
        np.array([70, 5, -45]),
        bounds=np.array([[30, 90], [0, 30], [-90, 0]]),
        standard_name="latitude",
        units="degrees",
    )
    lon = iris.coords.DimCoord(
        np.array([45, 135, 225, 315]),
        bounds=np.array([[0, 90], [90, 180], [180, 270], [270, 360]]),
        standard_name="longitude",
        units="degrees",
        circular=True,
    )
    write_surface_frac_file(
        TEST_SFTLF_PATH, lat, lon, "land_area_fraction", "sftlf", "%"
    )
    write_surface_frac_file(
        TEST_SFTOF_PATH, lat, lon, "sea_area_fraction", "sftof", "%", inverse=True
    )
    write_area_file(TEST_AREACEALLA_PATH, lat, lon, "cell_area", "areacella", "m^2")
    write_area_file(TEST_AREACEALLO_PATH, lat, lon, "cell_area", "areacello", "m^2")
    write_data_file(
        TEST_RSDT_PATH,
        lat,
        lon,
        "toa_incoming_shortwave_flux",
        "rsdt",
        "W m-2",
        {"modeling_realm": "atmos"},
    )
    write_data_file(
        TEST_GPP_PATH,
        lat,
        lon,
        "gross_primary_productivity_of_carbon",
        "gpp",
        "kg m-2 s-1",
        {"modeling_realm": "land"},
    )
    write_data_file(
        TEST_CSOILFAST_PATH,
        lat,
        lon,
        "fast_soil_pool_carbon_content",
        "cSoilFast",
        "kg m-2",
        {"modeling_realm": "land"},
    )
    write_data_file(
        TEST_HFDS_PATH,
        lat,
        lon,
        "surface_downward_heat_flux_in_sea_water",
        "hfds",
        "W m-2",
        {"modeling_realm": "ocean"},
    )


def write_surface_frac_file(
    write_path, lat, lon, standard_name, var_name, units, inverse=False
):
    data = SURFACE_FRACS if not inverse else 100 - SURFACE_FRACS
    cube = iris.cube.Cube(
        data,
        standard_name=standard_name,
        var_name=var_name,
        units=units,
        dim_coords_and_dims=[(lat, 0), (lon, 1)],
    )
    save_cube_in_path(cube, write_path)


def write_area_file(write_path, lat, lon, standard_name, var_name, units):
    cube = iris.cube.Cube(
        AREA_WEIGHTS,
        standard_name=standard_name,
        var_name=var_name,
        units=units,
        dim_coords_and_dims=[(lat, 0), (lon, 1)],
    )
    save_cube_in_path(cube, write_path)


def write_data_file(write_path, lat, lon, standard_name, var_name, units, attributes):
    time = iris.coords.DimCoord(
        np.array([15.5, 45, 74.5]),
        standard_name="time",
        units="days since 1850-01-01 00:00:00",
    )
    time.guess_bounds()

    cube = iris.cube.Cube(
        RAW_DATA,
        standard_name=standard_name,
        var_name=var_name,
        units=units,
        dim_coords_and_dims=[(time, 0), (lat, 1), (lon, 2)],
    )
    cube.attributes = attributes
    save_cube_in_path(cube, write_path)


def save_cube_in_path(cube, write_path):
    dir_to_save = os.path.dirname(write_path)
    if not os.path.isdir(dir_to_save):
        os.makedirs(dir_to_save)

    iris.save(cube, write_path)



"""
Helper tools for citing `Coupled Model Intercomparison Project <https://www.wcrp-climate.org/wgcm-cmip>`_ data
"""
import os.path
import warnings

import pandas as pd
import tqdm.autonotebook as tqdman

from ..errors import NonStandardLicenseError
from ..iris_cube_wrappers import CMIP6OutputCube, MarbleCMIP5Cube
from ..retractions import _get_dependencies
from .cmip5 import (
    _determine_cmip5_modelling_group,
    _get_cmip5_citation_key,
    check_license_cmip5,
    get_cmip5_bibtex_citation,
)
from .cmip6 import (
    _get_cmip6_citation_key,
    check_license_cmip6,
    get_cmip6_bibtex_citation,
)


def check_licenses(scmruns):
    """
    Check datasets for non-standard licenses

    Non-standard licenses result in a warning

    Parameters
    ----------
    scmruns : list of :obj:`scmdata.ScmRun`
        Datasets to check the licenses of

    Returns
    -------
    list of :obj:`scmdata.ScmRun`
        Datasets with non-standard licenses
    """
    non_standard = []
    for scmrun in tqdman.tqdm(scmruns):
        mip_era = scmrun.get_unique_meta("mip_era", no_duplicates=True)

        try:
            if mip_era == "CMIP5":
                check_license_cmip5(scmrun)
            else:
                check_license_cmip6(scmrun)

        except NonStandardLicenseError as e:
            non_standard.append(scmrun)

            climate_model = scmrun.get_unique_meta("climate_model", no_duplicates=True)
            warnings.warn(
                "Non-standard license for {}\n"
                "License terms:\n\t{}".format(climate_model, e)
            )

    return non_standard


def get_citation_tables(database):
    """
    Get citation tables for a given set of CMIP data

    Parameters
    ----------
    database : list of :obj:`ScmRun`
        Set of CMIP data for which we want to create citation tables

    Returns
    -------
    dict of str: Union[List, pd.DataFrame]
        Dictionary containing the citation table and bibtex references
        for each MIP era used in ``database``

    Raises
    ------
    ValueError
        Any :obj:`ScmRun` in ``database`` has a ``mip_era`` other than "CMIP5"
        or "CMIP6"
    """
    cmip5_references = {
        "Climate Model": [],
        "Scenario": [],
        "Modelling Group": [],
        "Reference": [],
    }
    cmip5_bibtex = []

    cmip6_references = {
        "Climate Model": [],
        "Scenario": [],
        "Modelling Group": [],
        "Reference": [],
    }
    cmip6_bibtex = []

    for scmrun in tqdman.tqdm(database, desc="Datasets"):
        mip_era = scmrun.get_unique_meta("mip_era", no_duplicates=True)
        if mip_era not in ["CMIP5", "CMIP6"]:
            raise ValueError(
                "Cannot generate citation table for mip_era: {}".format(mip_era)
            )

        dependencies = _get_dependencies(scmrun.metadata)

        if not dependencies:
            raise ValueError("Could not find any dependencies for {}".format(scmrun))

        for dependency in tqdman.tqdm(
            dependencies, desc="Dependencies", total=len(dependencies), leave=False
        ):
            if mip_era == "CMIP5":
                cmip5_references, cmip5_bibtex = _process_cmip5_dependency(
                    dependency, cmip5_references, cmip5_bibtex
                )

            else:  # CMIP6 thanks to check above
                institute_id = set(
                    [v for k, v in scmrun.metadata.items() if "institution_id" in k]
                )

                if len(institute_id) != 1:  # pragma: no cover
                    raise AssertionError(
                        "More than one institute? {}".format(institute_id)
                    )

                institute_id = list(institute_id)[0]

                cmip6_references, cmip6_bibtex = _process_cmip6_dependency(
                    dependency, cmip6_references, cmip6_bibtex, institute_id
                )

    cmip5_references = (
        pd.DataFrame(cmip5_references)
        .set_index(["Modelling Group", "Climate Model", "Scenario"])
        .sort_index()
        .drop_duplicates()
    )
    cmip5_bibtex = list(set(cmip5_bibtex))

    cmip6_references = (
        pd.DataFrame(cmip6_references)
        .set_index(["Modelling Group", "Climate Model", "Scenario"])
        .sort_index()
        .drop_duplicates()
    )
    cmip6_bibtex = list(set(cmip6_bibtex))

    return {
        "cmip5_references": cmip5_references,
        "cmip5_bibtex": cmip5_bibtex,
        "cmip6_references": cmip6_references,
        "cmip6_bibtex": cmip6_bibtex,
    }


def _process_cmip5_dependency(dependency, references, bibtexs):
    helper = MarbleCMIP5Cube()

    path_bits = helper.process_path(os.path.dirname(dependency))
    climate_model = path_bits["model"]
    scenario = path_bits["experiment"]
    citation_key = _get_cmip5_citation_key(climate_model, scenario)

    ref = r"\citet{" + citation_key + "}"

    if ref not in references["Reference"]:
        model_group = _determine_cmip5_modelling_group(climate_model)

        references["Climate Model"].append(climate_model)
        references["Scenario"].append(scenario)
        references["Modelling Group"].append(model_group)
        references["Reference"].append(ref)

        bib_cite = get_cmip5_bibtex_citation(climate_model, scenario)

        bibtexs.append(bib_cite)

    return references, bibtexs


def _process_cmip6_dependency(dependency, references, bibtexs, institute_id):
    helper = CMIP6OutputCube()

    path_bits = helper.process_path(os.path.dirname(dependency))

    activity_id = path_bits["activity_id"]
    climate_model = path_bits["source_id"]
    scenario = path_bits["experiment_id"]

    citation_key = _get_cmip6_citation_key(
        path_bits["mip_era"], activity_id, institute_id, climate_model, scenario,
    )
    ref = r"\citet{" + citation_key + "}"

    if ref not in references["Reference"]:
        references["Climate Model"].append(climate_model)
        references["Scenario"].append(scenario)
        references["Modelling Group"].append(institute_id)
        references["Reference"].append(ref)

        bib_cite = get_cmip6_bibtex_citation(
            path_bits["mip_era"],
            activity_id,
            institute_id,
            climate_model,
            scenario,
            version=path_bits["version"],
        )

        bibtexs.append(bib_cite)

    return references, bibtexs



"""
Utility functions for command-line relevant operations
"""
# TODO: clean up so usage of these bits and pieces is clearer
import datetime as dt
import logging
import os
import os.path
import re
import sys
from time import gmtime, strftime

import numpy as np
import pymagicc
import scmdata.units
import tqdm
from pymagicc.io import MAGICCData
from scmdata import df_append

from . import __version__
from .io import get_scmcube_helper, load_scmdataframe
from .iris_cube_wrappers import ScmCube

LOGGER = logging.getLogger("netcdf_scm")

# TODO: use openscm_units
_ureg = scmdata.units.ScmUnitRegistry()
"""
Unit registry for miscellaneous unit checking
"""
_ureg.add_standards()

_MAGICC_VARIABLE_MAP = {"tas": ("Surface Temperature", "SURFACE_TEMP")}
"""Mapping from CMOR variable names to MAGICC variables"""


def _init_logging(logger, params, out_filename=None):
    """
    Set up a logger

    All INFO messages and greater are written to stderr.
    If an ``out_filename`` is provided, all recorded log messages are also written to
    disk.

    # TODO: make level of logging customisable

    Parameters
    ----------
    logger : :obj:`logging.Logger`
        Logger to initialise

    params : list
        A list of key values to write at the start of the log

    out_filename : str
        Name of the log file which is written to disk
    """
    handlers = []
    if out_filename:
        h = logging.FileHandler(out_filename, "a")
        h.setLevel(logging.DEBUG)
        handlers.append(h)

    # Write logs to stderr
    h_stderr = logging.StreamHandler(sys.stderr)
    h_stderr.setLevel(logging.INFO)
    handlers.append(h_stderr)

    fmt = logging.Formatter(
        "{process} {asctime} {levelname}:{name}:{message}", style="{"
    )
    for h in handlers:
        if h.formatter is None:
            h.setFormatter(fmt)
        logger.addHandler(h)

    # use DEBUG as default for now
    logger.setLevel(logging.DEBUG)

    logging.captureWarnings(True)
    logger.info("netcdf-scm: %s", __version__)
    for k, v in params:
        logger.info("%s: %s", k, v)


def _get_timestamp():
    return strftime("%Y%m%d %H%M%S", gmtime())


def _make_path_if_not_exists(path_to_check):
    if not os.path.exists(path_to_check):
        LOGGER.info("Making output directory: %s", path_to_check)
        os.makedirs(path_to_check)


def _load_scm_cube(drs, dirpath, filenames):
    scmcube = get_scmcube_helper(drs)
    if len(filenames) == 1:
        scmcube.load_data_from_path(os.path.join(dirpath, filenames[0]))
    else:
        scmcube.load_data_in_directory(dirpath)

    return scmcube


def _set_crunch_contact_in_results(res, crunch_contact):
    for _, c in res.items():
        if "crunch_contact" in c.cube.attributes:
            LOGGER.warning(
                "Overwriting `crunch_contact` attribute"
            )  # pragma: no cover # emergency valve
        c.cube.attributes["crunch_contact"] = crunch_contact

    return res


def _get_outfile_dir(dpath, drs, dst):
    scmcube = get_scmcube_helper(drs)
    outfile_dir = dpath.replace(scmcube.process_path(dpath)["root_dir"], dst)
    _make_path_if_not_exists(outfile_dir)

    return outfile_dir


def _convert_units(openscmdf, target_units_specs):
    for variable in openscmdf["variable"].unique():
        if variable in target_units_specs["variable"].tolist():
            target_unit = target_units_specs[
                target_units_specs["variable"] == variable
            ]["unit"].values[0]
            current_unit = openscmdf.filter(variable=variable)["unit"].values[0]

            LOGGER.info(
                "Converting units of %s from %s to %s",
                variable,
                current_unit,
                target_unit,
            )

            target_length = _ureg(target_unit).dimensionality["[length]"]
            current_length = _ureg(current_unit).dimensionality["[length]"]
            if np.equal(target_length - current_length, 2):
                openscmdf = _take_area_sum(openscmdf, current_unit)

            openscmdf = openscmdf.convert_unit(target_unit, variable=variable)

    return openscmdf


def _find_dirs_meeting_func(src, check_func):
    matching_dirs = []
    failures = False
    LOGGER.info("Finding directories with files")
    tqdm_kwargs = {
        "desc": "Walking through directories and applying `check_func`",
        "leave": True,
    }
    for dirpath, _, filenames in tqdm.tqdm(os.walk(src), **tqdm_kwargs):
        LOGGER.debug("Entering %s", dirpath)
        if filenames:
            try:
                if check_func(dirpath):
                    matching_dirs.append((dirpath, filenames))
            except Exception as e:  # pylint:disable=broad-except
                LOGGER.error(
                    "Directory checking failed on %s with error %s", dirpath, e
                )
                failures = True

    LOGGER.info("Found %s directories with files", len(matching_dirs))
    return matching_dirs, failures


def _match_regexp_list(list_to_check, regexp):
    regexp_compiled = re.compile(regexp)

    return [v for v in list_to_check if regexp_compiled.match(v)]


def _skip_file(out_file, force):
    if not force and os.path.isfile(out_file):
        LOGGER.info("Skipped (already exists, not overwriting) %s", out_file)
        return True

    if os.path.isfile(out_file):
        os.remove(out_file)

    return False


def _get_path_bits(inpath, drs):
    helper = get_scmcube_helper(drs)
    return helper.process_path(os.path.dirname(inpath))


def _get_id_in_path(path_id, fullpath, drs):
    # TODO: change to _get_path_bits(fullpath, drs)[path_id]
    helper = get_scmcube_helper(drs)
    return helper.process_path(os.path.dirname(fullpath))[path_id]


def _get_timestamp_str(fullpath, drs):
    # TODO: rename to _get_timestamp_str_from_path
    # TODO: change to _get_id_in_path("timestamp_str", fullpath, drs)
    helper = get_scmcube_helper(drs)
    filename_bits = helper._get_timestamp_bits_from_filename(  # pylint:disable=protected-access
        os.path.basename(fullpath)
    )
    return filename_bits["timestamp_str"]


def _get_meta(inscmdf, meta_col, expected_unique=True):
    # TODO: remove and instead use scmdataframe.get_unique_meta(raise=True) or whatever it is
    vals = inscmdf[meta_col].unique()
    if expected_unique:
        if len(vals) != 1:
            raise AssertionError("{} is not unique: {}".format(meta_col, vals))
        return vals[0]

    return vals


def _write_ascii_file(  # pylint:disable=too-many-arguments
    openscmdf,
    metadata,
    header,
    outfile_dir,
    fnames,
    force,
    out_format,
    drs,
    prefix=None,
):
    if out_format in ("mag-files",):
        _write_mag_file(
            openscmdf, metadata, header, outfile_dir, fnames, force, prefix, drs,
        )
    elif out_format in (
        "mag-files-average-year-start-year",
        "mag-files-average-year-mid-year",
        "mag-files-average-year-end-year",
        "mag-files-point-start-year",
        "mag-files-point-mid-year",
        "mag-files-point-end-year",
    ):
        _write_mag_file_with_operation(
            openscmdf,
            metadata,
            header,
            outfile_dir,
            fnames,
            force,
            out_format,
            prefix,
            drs,
        )
    elif out_format in ("magicc-input-files",):
        _write_magicc_input_file(
            openscmdf, metadata, header, outfile_dir, fnames, force, prefix,
        )
    elif out_format in (
        "magicc-input-files-average-year-start-year",
        "magicc-input-files-average-year-mid-year",
        "magicc-input-files-average-year-end-year",
        "magicc-input-files-point-start-year",
        "magicc-input-files-point-mid-year",
        "magicc-input-files-point-end-year",
    ):
        _write_magicc_input_file_with_operation(
            openscmdf, metadata, header, outfile_dir, fnames, force, out_format, prefix,
        )
    else:
        raise AssertionError("how did we get here?")  # pragma: no cover


def _write_mag_file(  # pylint:disable=too-many-arguments,too-many-locals
    openscmdf, metadata, header, outfile_dir, fnames, force, prefix, drs
):
    ts = openscmdf.timeseries()

    src_time_points = ts.columns

    time_id = "{:04d}{:02d}-{:04d}{:02d}".format(
        src_time_points[0].year,
        src_time_points[0].month,
        src_time_points[-1].year,
        src_time_points[-1].month,
    )
    old_time_id = _get_timestamp_str(fnames[0], drs)

    out_file_base = fnames[0].replace(old_time_id, time_id)
    if prefix is not None:
        out_file_base = "{}_{}".format(prefix, out_file_base)

    out_file = os.path.join(outfile_dir, out_file_base)
    out_file = "{}.MAG".format(os.path.splitext(out_file)[0])

    if _skip_file(out_file, force):
        return

    writer = MAGICCData(openscmdf)
    writer["todo"] = "SET"
    _check_timesteps_are_monthly(writer)

    writer.metadata = metadata
    writer.metadata["timeseriestype"] = "MONTHLY"
    writer.metadata["header"] = header

    LOGGER.info("Writing file to %s", out_file)
    writer.write(out_file, magicc_version=7)


def _check_timesteps_are_monthly(scmdf):
    time_steps = scmdf["time"][1:].values - scmdf["time"][:-1].values
    step_upper = np.timedelta64(32, "D")  # pylint:disable=too-many-function-args
    step_lower = np.timedelta64(28, "D")  # pylint:disable=too-many-function-args
    if any((time_steps > step_upper) | (time_steps < step_lower)):
        raise ValueError(
            "Please raise an issue at "
            "gitlab.com/netcdf-scm/netcdf-scm/issues "
            "to discuss how to handle non-monthly data"
        )


def _write_mag_file_with_operation(  # pylint:disable=too-many-arguments
    openscmdf, metadata, header, outfile_dir, fnames, force, out_format, prefix, drs,
):  # pylint:disable=too-many-locals
    if len(fnames) > 1:
        raise AssertionError(
            "more than one file to wrangle?"
        )  # pragma: no cover # emergency valve

    ts = openscmdf.timeseries()

    src_time_points = ts.columns
    original_years = ts.columns.map(lambda x: x.year).unique()

    time_id = "{:04d}-{:04d}".format(src_time_points[0].year, src_time_points[-1].year)
    old_time_id = _get_timestamp_str(fnames[0], drs)

    out_file_base = fnames[0].replace(old_time_id, time_id)
    if prefix is not None:
        out_file_base = "{}_{}".format(prefix, out_file_base)

    out_file = os.path.join(outfile_dir, out_file_base)
    out_file = "{}.MAG".format(os.path.splitext(out_file)[0])

    if _skip_file(out_file, force):
        return

    writer = MAGICCData(_do_timeseriestype_operation(openscmdf, out_format)).filter(
        year=original_years
    )

    writer["todo"] = "SET"
    writer.metadata = metadata
    writer.metadata["timeseriestype"] = (
        out_format.replace("mag-files-", "").replace("-", "_").upper()
    )

    writer.metadata["header"] = header

    LOGGER.info("Writing file to %s", out_file)
    writer.write(out_file, magicc_version=7)


def _do_timeseriestype_operation(openscmdf, out_format):
    if out_format.endswith("average-year-start-year"):
        out = openscmdf.time_mean("AS")

    elif out_format.endswith("average-year-mid-year"):
        out = openscmdf.time_mean("AC")

    elif out_format.endswith("average-year-end-year"):
        out = openscmdf.time_mean("A")

    elif out_format.endswith("point-start-year"):
        out = openscmdf.resample("AS")

    elif out_format.endswith("point-mid-year"):
        out_time_points = [
            dt.datetime(y, 7, 1)
            for y in range(
                openscmdf["time"].min().year, openscmdf["time"].max().year + 1
            )
        ]
        out = openscmdf.interpolate(target_times=out_time_points)

    elif out_format.endswith("point-end-year"):
        out = openscmdf.resample("A")

    else:  # pragma: no cover # emergency valve
        raise NotImplementedError("Do not recognise out_format: {}".format(out_format))

    if out.timeseries().shape[1] == 1:
        error_msg = "We cannot yet write `{}` if the output data will have only one timestep".format(
            out_format
        )
        raise ValueError(error_msg)

    return out


def _write_magicc_input_file(  # pylint:disable=too-many-arguments
    openscmdf, metadata, header, outfile_dir, fnames, force, prefix
):
    if len(fnames) > 1:
        raise AssertionError(
            "more than one file to wrangle?"
        )  # pragma: no cover # emergency valve

    _write_magicc_input_files(
        openscmdf, outfile_dir, force, metadata, header, "MONTHLY", prefix,
    )


def _write_magicc_input_file_with_operation(  # pylint:disable=too-many-arguments
    openscmdf, metadata, header, outfile_dir, fnames, force, out_format, prefix,
):
    if len(fnames) > 1:
        raise AssertionError(
            "more than one file to wrangle?"
        )  # pragma: no cover # emergency valve

    ts = openscmdf.timeseries()

    original_years = ts.columns.map(lambda x: x.year).unique()

    openscmdf = _do_timeseriestype_operation(openscmdf, out_format).filter(
        year=original_years
    )

    _write_magicc_input_files(
        openscmdf,
        outfile_dir,
        force,
        metadata,
        header,
        out_format.replace("magicc-input-files-", "").replace("-", "_").upper(),
        prefix,
    )


def _write_magicc_input_files(  # pylint:disable=too-many-arguments,too-many-locals
    openscmdf, outfile_dir, force, metadata, header, timeseriestype, prefix,
):
    try:
        var_to_write = openscmdf["variable"].unique()[0]
        variable_abbreviations = {
            "filename": var_to_write,
            "magicc_name": _MAGICC_VARIABLE_MAP[var_to_write][0],
            "magicc_internal_name": _MAGICC_VARIABLE_MAP[var_to_write][1],
        }
    except KeyError:
        raise KeyError(
            "I don't know which MAGICC variable to use for input `{}`".format(
                var_to_write
            )
        )

    region_filters = {
        "FOURBOX": [
            "World|Northern Hemisphere|Land",
            "World|Southern Hemisphere|Land",
            "World|Northern Hemisphere|Ocean",
            "World|Southern Hemisphere|Ocean",
        ],
        "GLOBAL": ["World"],
    }
    for region_key, regions_to_keep in region_filters.items():
        out_file_base = (
            ("{}_{}_{}_{}_{}_{}.IN")
            .format(
                variable_abbreviations["filename"],
                openscmdf["scenario"].unique()[0],
                openscmdf["climate_model"].unique()[0],
                openscmdf["member_id"].unique()[0],
                region_key,
                variable_abbreviations["magicc_internal_name"],
            )
            .upper()
        )
        if prefix is not None:
            out_file_base = "{}_{}".format(prefix, out_file_base)

        out_file = os.path.join(outfile_dir, out_file_base,)

        if _skip_file(out_file, force):
            return

        writer = MAGICCData(openscmdf).filter(region=regions_to_keep)
        writer["todo"] = "SET"
        writer["variable"] = variable_abbreviations["magicc_name"]
        writer.metadata = metadata
        writer.metadata["header"] = header
        writer.metadata["timeseriestype"] = timeseriestype

        LOGGER.info("Writing file to %s", out_file)
        writer.write(out_file, magicc_version=7)


def _get_openscmdf_metadata_header(
    fnames, dpath, target_units_specs, wrangle_contact, out_format
):
    if len(fnames) > 1:
        raise AssertionError(
            "more than one file to wrangle?"
        )  # pragma: no cover # emergency valve

    openscmdf = df_append([load_scmdataframe(os.path.join(dpath, f)) for f in fnames])
    if openscmdf.timeseries().shape[1] == 1:
        error_msg = "We cannot yet write `{}` if the output data has only one timestep".format(
            out_format
        )
        raise ValueError(error_msg)

    if target_units_specs is not None:
        openscmdf = _convert_units(openscmdf, target_units_specs)

    metadata = openscmdf.metadata
    header = _get_openscmdf_header(
        wrangle_contact, metadata["crunch_netcdf_scm_version"]
    )

    return openscmdf, metadata, header


def _get_openscmdf_header(contact, netcdf_scm_version):
    header = (
        "Date: {}\n"
        "Contact: {}\n"
        "Source data crunched with: netCDF-SCM v{}\n"
        "File written with: pymagicc v{} (more info at "
        "github.com/openclimatedata/pymagicc)\n".format(
            dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            contact,
            netcdf_scm_version,
            pymagicc.__version__,
        )
    )

    return header


def _take_area_sum(openscmdf, current_unit):
    converted_ts = []

    for region, df in openscmdf.timeseries().groupby("region"):
        rkey = ScmCube._convert_region_to_area_key(  # pylint:disable=protected-access
            region
        )
        for k, v in openscmdf.metadata.items():
            if "{} (".format(rkey) in k:
                unit = k.split("(")[-1].split(")")[0]
                conv_factor = v * _ureg(unit)

                converted_region = df * v
                converted_region = converted_region.reset_index()
                converted_region["unit"] = str(
                    (1 * _ureg(current_unit) * conv_factor).units
                )
                converted_ts.append(converted_region)
                break

    converted_ts = df_append(converted_ts)
    converted_ts.metadata = openscmdf.metadata
    return converted_ts



import re

import pandas as pd
import pytest
from base_normalisation_integration import NormaliserIntegrationTester

from netcdf_scm.normalisation import AfterBranchTimeMean
from netcdf_scm.stitching import get_branch_time


class TestRunningMean(NormaliserIntegrationTester):
    tclass = AfterBranchTimeMean

    @pytest.mark.parametrize("member_id", ("r1i1p1f1", "r10i1p1f1"))
    @pytest.mark.parametrize("activity_id", ("ScenarioMIP", "RFMIP"))
    def test_get_reference_values(
        self,
        historical_data_normalisation,
        picontrol_data_normalisation,
        member_id,
        activity_id,
    ):
        picontrol_branch_time = get_branch_time(historical_data_normalisation)

        ref_mean = (
            picontrol_data_normalisation.filter(
                year=range(picontrol_branch_time.year, picontrol_branch_time.year + 31)
            )
            .timeseries()
            .mean(axis=1)
        )
        expected = pd.concat(
            [ref_mean] * historical_data_normalisation["time"].shape[0], axis=1
        )
        expected.columns = historical_data_normalisation["time"]

        expected = expected.reset_index()
        idx = historical_data_normalisation.meta.columns.tolist()
        for c in idx:
            if c in ["region", "variable", "unit"]:
                continue

            expected[c] = historical_data_normalisation.get_unique_meta(
                c, no_duplicates=True
            )

        expected = expected.set_index(idx)

        picontrol_data_to_use = picontrol_data_normalisation.copy()
        picontrol_data_to_use.set_meta(member_id, "member_id")
        picontrol_data_to_use.set_meta(activity_id, "activity_id")

        res = self.tclass().get_reference_values(
            historical_data_normalisation, picontrol_data_to_use, picontrol_branch_time,
        )

        pd.testing.assert_frame_equal(res, expected, check_like=True)

    def test_get_reference_values_not_enough_data_error(
        self, historical_data_normalisation, picontrol_data_normalisation
    ):
        picontrol_branch_time = get_branch_time(historical_data_normalisation)
        pi_data_to_use = picontrol_data_normalisation.filter(
            year=range(picontrol_branch_time.year, picontrol_branch_time.year + 10)
        )

        error_msg = re.compile(
            ".*Only `084101` to `085012` is available after the branching time `084101` in piControl "
            "data in "
            "{}".format(pi_data_to_use.metadata["netcdf-scm crunched file"])
        )

        with pytest.raises(ValueError, match=error_msg):
            self.tclass().get_reference_values(
                historical_data_normalisation, pi_data_to_use, picontrol_branch_time,
            )



import logging
import re
from unittest.mock import MagicMock, PropertyMock, call, patch

import iris
import numpy as np
import pytest
from conftest import create_sftlf_cube

from netcdf_scm.iris_cube_wrappers import ScmCube
from netcdf_scm.weights import (
    DEFAULT_REGIONS,
    WEIGHTS_FUNCTIONS_WITHOUT_AREA_WEIGHTING,
    AreaSurfaceFractionWeightCalculator,
    AreaWeightCalculator,
    get_default_sftlf_cube,
    get_land_weights,
    get_ocean_weights,
    get_weights_for_area,
)


def create_dummy_cube_from_lat_lon_points(lat_pts, lon_pts):
    lat = iris.coords.DimCoord(lat_pts, standard_name="latitude", units="degrees")
    lon = iris.coords.DimCoord(
        lon_pts, standard_name="longitude", units="degrees", circular=True
    )

    cube = iris.cube.Cube(
        np.full((len(lat_pts), len(lon_pts)), 0),
        dim_coords_and_dims=[(lat, 0), (lon, 1)],
    )

    return cube


class _WeightRetrievalTester:
    # cube weight calculator
    cwc = None

    @patch("netcdf_scm.weights.get_land_weights")
    @patch("netcdf_scm.weights.get_nh_weights")
    def test_get_scm_masks(self, mock_nh_weights, mock_land_weights, test_all_cubes):
        tsftlf_cube = "mocked 124"

        land_weights = np.array([[1, 0, 0, 1], [1, 0, 1, 0], [1, 1, 0, 1]])
        mock_land_weights.return_value = land_weights

        nh_weights = np.array([[1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 0, 0]])
        mock_nh_weights.return_value = nh_weights

        nh_land_weights = np.array([[1, 0, 0, 1], [1, 0, 1, 0], [0, 0, 0, 0]])
        # check our logic while we're here
        np.testing.assert_array_equal(nh_weights * land_weights, nh_land_weights)

        area_weights = test_all_cubes.get_area_weights()
        expected = {
            k: self._expected_area_non_area_weights_combination(area_weights, v)
            for k, v in {
                "World": np.full(nh_weights.shape, 1),
                "World|Northern Hemisphere|Land": nh_land_weights,
                "World|Southern Hemisphere|Land": (1 - nh_weights) * land_weights,
                "World|Northern Hemisphere|Ocean": nh_weights * (1 - land_weights),
                "World|Southern Hemisphere|Ocean": (
                    (1 - nh_weights) * (1 - land_weights)
                ),
                "World|Land": land_weights,
                "World|Ocean": 1 - land_weights,
                "World|Northern Hemisphere": nh_weights,
                "World|Southern Hemisphere": 1 - nh_weights,
            }.items()
        }

        with patch.dict(
            WEIGHTS_FUNCTIONS_WITHOUT_AREA_WEIGHTING,
            {
                "World|Northern Hemisphere": mock_nh_weights,
                "World|Land": mock_land_weights,
            },
        ):
            weighter = self.cwc(test_all_cubes, sftlf_cube=tsftlf_cube)
            result = weighter.get_weights(DEFAULT_REGIONS)

        for label, array in expected.items():
            np.testing.assert_array_equal(array, result[label])
        mock_land_weights.assert_called_with(
            weighter, test_all_cubes, sftlf_cube=tsftlf_cube
        )
        mock_nh_weights.assert_called_with(
            weighter, test_all_cubes, sftlf_cube=tsftlf_cube
        )

    @pytest.mark.parametrize("with_bounds", [True, False])
    @patch("netcdf_scm.weights.get_nh_weights")
    def test_get_scm_masks_no_land_available(
        self, mock_nh_weights, with_bounds, test_all_cubes, caplog
    ):
        caplog.set_level(logging.WARNING, logger="netcdf_scm.iris_cube_wrappers")
        caplog.set_level(logging.DEBUG, logger="netcdf_scm.weights")
        test_all_cubes.get_metadata_cube = MagicMock(side_effect=OSError)

        nh_weights = np.array([[1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 0, 0]])
        mock_nh_weights.return_value = nh_weights
        default_sftlf_cube = get_default_sftlf_cube()
        default_sftlf_cube = default_sftlf_cube.regrid(
            # AreaWeighted() in future but too slow now
            test_all_cubes.cube,
            iris.analysis.Linear(),
        )
        expected_land_weights = default_sftlf_cube.data
        # normalise to 1
        expected_land_weights = expected_land_weights / expected_land_weights.max()

        area_weights = test_all_cubes.get_area_weights()
        expected = {
            k: self._expected_area_non_area_weights_combination(area_weights, v)
            for k, v in {
                "World": np.full(nh_weights.shape, 1),
                "World|Northern Hemisphere": nh_weights,
                "World|Northern Hemisphere|Land": nh_weights * expected_land_weights,
                "World|Southern Hemisphere": 1 - nh_weights,
            }.items()
        }
        expected_warn = (
            "Land surface fraction (sftlf) data not available, using default instead"
        )
        norm_message = (
            "sftlf data max is 100.0, dividing by 100.0 to convert units to fraction"
        )
        with patch.dict(
            WEIGHTS_FUNCTIONS_WITHOUT_AREA_WEIGHTING,
            {"World|Northern Hemisphere": mock_nh_weights},
        ):
            if not with_bounds:
                test_all_cubes.lat_dim.bounds = None
                test_all_cubes.lon_dim.bounds = None

            weighter = self.cwc(test_all_cubes)
            result = weighter.get_weights(expected.keys())

        if not with_bounds:
            assert len(caplog.messages) == 5
            assert (
                caplog.messages[0]
                == "Couldn't find/use areacell_cube, falling back to iris.analysis.cartography.area_weights"
            )
            assert caplog.messages[0] == caplog.messages[1]
            assert caplog.messages[2] == "Guessing latitude and longitude bounds"
            assert caplog.messages[3] == expected_warn
            assert caplog.messages[4] == norm_message

        else:
            assert len(caplog.messages) == 4
            assert (
                caplog.messages[0]
                == "Couldn't find/use areacell_cube, falling back to iris.analysis.cartography.area_weights"
            )
            assert caplog.messages[0] == caplog.messages[1]
            assert caplog.messages[2] == expected_warn
            assert caplog.messages[3] == norm_message

        for label, array in expected.items():
            np.testing.assert_array_equal(array, result[label])

        mock_nh_weights.assert_called_with(weighter, test_all_cubes)

    @pytest.mark.parametrize("input_format", ["scmcube", None])
    @pytest.mark.parametrize("sftlf_var", ["sftlf", "sftlf_other"])
    @patch(
        "netcdf_scm.iris_cube_wrappers.ScmCube.surface_fraction_var",
        new_callable=PropertyMock,
    )
    def test_get_land_weights(
        self, mock_surface_fraction_var, test_all_cubes, input_format, sftlf_var
    ):
        sftlf_cube = create_sftlf_cube(test_all_cubes.__class__)
        mock_surface_fraction_var.return_value = sftlf_var
        original_data = sftlf_cube.cube.data

        test_all_cubes.get_metadata_cube = MagicMock(return_value=sftlf_cube)

        test_land_fraction_input = sftlf_cube if input_format == "scmcube" else None

        weighter = self.cwc(test_all_cubes)
        result = get_land_weights(
            weighter, test_all_cubes, sftlf_cube=test_land_fraction_input
        )

        expected = original_data / 100
        np.testing.assert_array_equal(result, expected)
        # Check that the sftlf meta cube is always registered
        test_all_cubes.get_metadata_cube.assert_called_with(
            sftlf_var, cube=test_land_fraction_input
        )

    @pytest.mark.parametrize("surface_frac_var", ["sftlf", "sftof"])
    @patch(
        "netcdf_scm.iris_cube_wrappers.ScmCube.netcdf_scm_realm",
        new_callable=PropertyMock,
    )
    def test_get_land_weights_shape_errors(
        self, mock_netcdf_scm_realm, surface_frac_var, test_all_cubes
    ):
        surface_frac_cube = create_sftlf_cube(test_all_cubes.__class__)
        error_msg = re.escape(
            "the {} cube data must be the same shape as the "
            "cube's longitude-latitude grid".format(surface_frac_var)
        )

        wrong_shape_data = np.array([[1, 2], [3, 4]])
        surface_frac_cube.cube = iris.cube.Cube(data=wrong_shape_data)

        mock_netcdf_scm_realm.return_value = (
            "land" if surface_frac_var == "sftlf" else "ocean"
        )

        weighter = self.cwc(test_all_cubes)
        with pytest.raises(AssertionError, match=error_msg):
            if surface_frac_var == "sftlf":
                get_land_weights(weighter, test_all_cubes, sftlf_cube=surface_frac_cube)
            else:
                get_ocean_weights(
                    weighter, test_all_cubes, sftof_cube=surface_frac_cube
                )

        test_all_cubes.get_metadata_cube = MagicMock(return_value=surface_frac_cube)
        with pytest.raises(AssertionError, match=error_msg):
            if surface_frac_var == "sftlf":
                get_land_weights(weighter, test_all_cubes, sftlf_cube=None)
            else:
                get_ocean_weights(weighter, test_all_cubes, sftof_cube=None)

    def test_nao_weights(self, test_all_cubes):
        sftlf_cube = create_sftlf_cube(test_all_cubes.__class__)
        weighter = self.cwc(test_all_cubes, sftlf_cube=sftlf_cube)
        result = weighter.get_weights_array_without_area_weighting(
            "World|North Atlantic Ocean"
        )

        expected = np.array([[0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 0]]) * (
            (100 - sftlf_cube.cube.data) / 100  # surface fraction normalised to 1
        )

        np.testing.assert_array_equal(result, expected)

    def test_elnino_weights(self, test_all_cubes):
        sftlf_cube = create_sftlf_cube(test_all_cubes.__class__)
        weighter = self.cwc(test_all_cubes, sftlf_cube=sftlf_cube)
        result = weighter.get_weights_array_without_area_weighting("World|El Nino N3.4")
        # 5N-5S, 190E-240E
        expected = np.array([[0, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 0]]) * (
            (100 - sftlf_cube.cube.data) / 100  # surface fraction normalised to 1
        )

        np.testing.assert_array_equal(result, expected)

    @pytest.mark.parametrize(
        "lat_pts,lon_pts,expected",
        [
            (  # nothing within bounds, raises Error
                np.array([-60, -1, 80]),
                np.array([45, 135, 225, 315]),
                "error",
            ),
            (  # nothing within bounds, raises Error
                np.array([-60, 10, 80]),
                np.array([45, 135, 225, 279]),
                "error",
            ),
            (  # nothing within bounds negative co-ord, raises Error
                np.array([-60, -1, 80]),
                np.array([-135, -45, 45, 135]),
                "error",
            ),
            (  # edge of bound included
                np.array([65, 0, -60]),
                np.array([45, 135, 225, 280]),
                np.array([[0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 0]]),
            ),
            (  # edge of bound included negative co-ord
                np.array([66, 0, -60]),
                np.array([-135, -80, 45, 135]),
                np.array([[0, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 0]]),
            ),
            (  # one within bounds
                np.array([80, 35, -70]),
                np.array([10, 30, 50, 135, 320]),
                np.array([[0, 0, 0, 0, 0], [0, 0, 0, 0, 1], [0, 0, 0, 0, 0]]),
            ),
            (  # one within bounds negative co-ord
                np.array([80, 35, -70]),
                np.array([-95, -40, 40, 135]),
                np.array([[0, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 0]]),
            ),
        ],
    )
    @pytest.mark.parametrize("query", [[0, -80, 65, 0], [0, 280, 65, 360]])
    def test_get_weights_for_area(
        self, test_all_cubes, query, lat_pts, lon_pts, expected
    ):
        regrid_cube = create_dummy_cube_from_lat_lon_points(lat_pts, lon_pts)
        test_all_cubes.cube = test_all_cubes.cube.regrid(
            regrid_cube, iris.analysis.Linear()
        )

        if isinstance(expected, str) and expected == "error":
            error_msg = re.compile("None of the cube's.*lie within the bounds.*")
            with pytest.raises(ValueError, match=error_msg):
                get_weights_for_area(*query)(None, test_all_cubes)
            return

        result = get_weights_for_area(*query)(None, test_all_cubes)
        np.testing.assert_array_equal(result, expected)

    def test_area_mask_wrapped_lons(self, test_all_cubes):
        result = get_weights_for_area(0, -80, 65, 0)(None, test_all_cubes)

        expected = np.array([[0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 0]])

        np.testing.assert_array_equal(result, expected)

    def test_get_masks_unknown_weights_warning(self, test_all_cubes, caplog):
        caplog.set_level(logging.WARNING, logger="netcdf_scm.iris_cube_wrappers")
        weighter = self.cwc(test_all_cubes)
        res = weighter.get_weights(["World", "junk"])

        np.testing.assert_allclose(res["World"], test_all_cubes.get_area_weights())

        assert len(caplog.messages) == 3
        assert (
            caplog.messages[0]
            == "Couldn't find/use areacell_cube, falling back to iris.analysis.cartography.area_weights"
        )
        assert caplog.records[0].levelname == "WARNING"
        assert (
            caplog.messages[1] == "Failed to create junk weights: Unknown weights: junk"
        )
        assert caplog.records[1].levelname == "WARNING"
        assert caplog.messages[0] == caplog.messages[2]
        assert caplog.records[0].levelname == caplog.records[2].levelname

    @pytest.mark.parametrize("exp_msg,cube_max", [(True, 100), (True, 53), (False, 1)])
    @pytest.mark.parametrize("surface_frac_var", ["sftlf", "sftof"])
    @patch(
        "netcdf_scm.iris_cube_wrappers.ScmCube.surface_fraction_var",
        new_callable=PropertyMock,
    )
    def test_get_scm_weights_surface_frac_bound_checks(
        self,
        mock_surface_fraction_var,
        exp_msg,
        cube_max,
        surface_frac_var,
        test_all_cubes,
        caplog,
    ):
        caplog.set_level(logging.DEBUG, logger="netcdf_scm.weights")
        tsurface_frac_cube = get_default_sftlf_cube().regrid(
            test_all_cubes.cube, iris.analysis.Linear()
        )
        tsftlf_cube_max = tsurface_frac_cube.data.max()
        assert np.isclose(tsftlf_cube_max, 100)
        if not np.equal(cube_max, 100):
            tsurface_frac_cube.data = tsurface_frac_cube.data / 100 * cube_max

        mock_surface_fraction_var.return_value = surface_frac_var
        test_all_cubes.get_metadata_cube = MagicMock()
        tsurface_frac_scmcube = ScmCube()
        tsurface_frac_scmcube.cube = tsurface_frac_cube

        tothermdata_scmcube = ScmCube()
        tothermdata_scmcube.cube = tsurface_frac_cube.copy()
        tothermdata_scmcube.cube.units = "m**2"

        def mock_meta_getter(metadata_variable, cube):
            if metadata_variable == test_all_cubes.surface_fraction_var:
                return tsurface_frac_scmcube

            return tothermdata_scmcube

        test_all_cubes.get_metadata_cube.side_effect = mock_meta_getter

        caplog.set_level(logging.INFO)
        weighter = self.cwc(test_all_cubes)
        weighter.get_weights(
            ["World|Land"] if surface_frac_var == "sftlf" else ["World|Ocean"]
        )

        test_all_cubes.get_metadata_cube.assert_has_calls(
            [call(test_all_cubes.surface_fraction_var, cube=None)]
        )

        if exp_msg:
            expected_msg = "{} data max is {}, dividing by {} to convert units to fraction".format(
                surface_frac_var,
                tsurface_frac_cube.data.max(),
                tsurface_frac_cube.data.max(),
            )

            assert len(caplog.messages) == 2
            assert caplog.messages[1] == expected_msg
            assert caplog.records[1].levelname == "DEBUG"
        else:
            assert len(caplog.messages) == 1

        assert "netCDF-SCM will treat the data as `atmosphere`" in caplog.messages[0]

    @staticmethod
    def _expected_area_non_area_weights_combination(area_weights, non_area_weights):
        raise NotImplementedError


class TestAreaSurfaceFractionWeightCalculatorWeightRetrieval(_WeightRetrievalTester):
    cwc = AreaSurfaceFractionWeightCalculator

    @staticmethod
    def _expected_area_non_area_weights_combination(area_weights, non_area_weights):
        return area_weights * non_area_weights


class TestAreaWeightCalculatorWeightRetrieval(_WeightRetrievalTester):
    cwc = AreaWeightCalculator

    @staticmethod
    def _expected_area_non_area_weights_combination(area_weights, non_area_weights):
        non_area_weights_rounded = np.array(~np.equal(non_area_weights, 0)).astype(int)
        return area_weights * non_area_weights_rounded



import pytest

from netcdf_scm.cli_utils import _match_regexp_list


@pytest.mark.parametrize(
    "regexp",
    (
        ".*tas.*",
        ".*/piControl/.*mon.*/(tas|fLUC)/(gr|gn)/.*",
        ".*/(piControl|.*)/.*mon.*/(tas|fLUC)/(gr|gn)/.*",
        ".*(.*).*/(piControl|.*)/.*mon.*/(tas|fLUC)/(gr|gn)/.*",
        # # this test takes ages to run because of the multiple capturing groups
        # # so off by default
        # ".*(.*).*/(piControl|.*)/.*mon.*(.*).*/(tas|fLUC)/(gr|gn)/.*",
    ),
)
@pytest.mark.parametrize(
    "inp_to_test",
    (
        (
            "./netcdf-scm/tests/test-data/cmip6output/CMIP6/ScenarioMIP/BCC/BCC-CSM2-MR/ssp126/r1i1p1f1/Amon/tas/gn/v20190314",
            "./netcdf-scm/tests/test-data/cmip6output/CMIP6/ScenarioMIP/BCC/BCC-CSM2-MR/ssp126/r1i1p1f1/Emon/fLUC/gn/v20190314",
            "./netcdf-scm/tests/test-data/cmip6output/CMIP6/ScenarioMIP/BCC/BCC-CSM2-MR/ssp126/r1i1p1f1/Omon/hfds/gn/v20190314",
        ),
    ),
)
@pytest.mark.parametrize("inp_scale_factor", (1, 2, 3, 4))
@pytest.mark.benchmark(max_time=1.0, min_rounds=1)
def test_regexp_match(benchmark, regexp, inp_to_test, inp_scale_factor):
    """
    Toy test function to examine the impact on match speed of using regexp
    of different complexity. Maybe one day we can update _match_regexp_list so
    it can handle super complex regexp using something like
    https://stackoverflow.com/a/42789508
    """
    inp_to_test = inp_to_test * inp_scale_factor

    res = benchmark(_match_regexp_list, inp_to_test, regexp)

    assert not any(["hdfs" in v for v in res])
